#!/usr/bin/env python

"""A translation of a Scheme interpreter in Python, modeling the
metacircular evaluator in Ch. 4 of The Structure and Interpretation of
Computer Programs.  (http://www-mitpress.mit.edu/sicp/)

AUTHOR:

    Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

SYNOPSIS:

    It's Scheme.  *grin*  At least, quite a bit of Scheme.

    Scheme is a variant of Lisp, and it's used in many introductory CS
    courses.  It is pretty powerful despite its deceptive size.  For
    more information on Scheme, visit the Scheme home page:

        http://www.swiss.ai.mit.edu/projects/scheme

BUGS:

    One major deficiency of this implementation is that it doesn't
    support tail recursion.  I don't believe I can implement this
    properly without fiddling around with the 'compiler' package and
    primitive bytecodes.  I'll read up on SICP Chapter 5.5 and see if
    this is feasible.

    August 26, 2004: this is not true; I CAN get real tail call
    elimination working, but it'll take CPS and trampolining to get it
    working right.  I'll first modify the interpreter into CPS form,
    and then apply trampolining from there.... ok, done!
    

MISCELLANEOUS:

    This source code is for the public domain; I'm writing it just for
    the fun of it.  This is certainly not meant to be efficient or
    even useful.  *grin*

    My future goals is to make this usable enough to run something
    like the explicit-register simulator and compiler in SICP Ch. 5.
"""



import parser
import sys

import evaluator
import builtins
import analyzer
import analyzer_builtins

import prompt
import error
import expressions
######################################################################


class Interpreter:
    """Meant to be subclassed.
    subclasses should define _eval and _env in their constructors."""
    def eval(self, exp):
        return self.getEvaluator()(exp, self.getEnvironment())

    def getEvaluator(self): return self._eval
    def getEnvironment(self): return self._env


class RegularInterpreter(Interpreter):
    def __init__(self):
        self._eval = evaluator.eval
        self._env = builtins.setupEnvironment()


class AnalyzingInterpreter(Interpreter):
    def __init__(self):
        self._eval = analyzer.eval
        self._env = analyzer_builtins.setupEnvironment()                       




def repl(interp):
    print 'Welcome to PyScheme!  Type: (QUIT) to quit.\n'
    def promptCallback(s):
        try:
            print expressions.expressionToString(interp.eval(parser.parse(s)))
        except (error.SchemeError, parser.ParserError), e:
            print "ERROR (%s): %s" % (e.__class__, e)

    p = prompt.Prompt(name="PyScheme", callback=promptCallback,
                      quit_str=None)
    p.promptLoop()
            

def isNonInteractive():
    """Returns True if we want to run in non-interactive mode."""
    if len(sys.argv) >= 2:
        return True


def runNonInteractive(interp):
    interp.eval(parser.parse("(load \"%s\")" %
                             sys.argv[1].replace('"', '\\"')))


def chooseInterpreter():
    return AnalyzingInterpreter()
##    return RegularInterpreter()

if __name__ == '__main__':
    interp = chooseInterpreter()
    if isNonInteractive():
        runNonInteractive(interp)
    else:
        repl(interp)
